@@ -120,8 +120,8 @@ enum cgdbrc_option_kind {
     CGDBRC_EXECUTING_LINE_DISPLAY,
     CGDBRC_HLSEARCH,
     CGDBRC_IGNORECASE,
+    CGDBRC_SCROLLBACK_BUFFER_SIZE,
     CGDBRC_SELECTED_LINE_DISPLAY,
-    CGDBRC_SHOWDEBUGCOMMANDS,
     CGDBRC_SHOWMARKS,
     CGDBRC_SYNTAX,
     CGDBRC_TABSTOP,
@@ -151,7 +151,7 @@ struct cgdbrc_config_option {
         /* option_kind == CGDBRC_DISASM */
         /* option_kind == CGDBRC_HLSEARCH */
         /* option_kind == CGDBRC_IGNORECASE */
-        /* option_kind == CGDBRC_SHOWDEBUGCOMMANDS */
+        /* option_kind == CGDBRC_SCROLLBACK_BUFFER_SIZE */
         /* option_kind == CGDBRC_SHOWMARKS */
         /* option_kind == CGDBRC_TABSTOP */
         /* option_kind == CGDBRC_TIMEOUT */
