@@ -7,7 +7,6 @@
  *  UNSET: for future expansion
  *  BIND: for future expansion
  *  MACRO: for future expansion
- *  BOOLEAN: yes or no
  *  NUMBER: a series of digits
  *  IDENTIFIER: a valid identifier (letter followed by letters or numbers)
  *  COMMAND: a recognized command (for future expansion)
@@ -21,7 +20,6 @@ enum TOKENS {
     MACRO,
     SHELL,
     BANG,
-    BOOLEAN,
     NUMBER,
     IDENTIFIER,
     COMMAND,
@@ -29,7 +27,7 @@ enum TOKENS {
     EOL
 };
 
-/* yylex: retreive the next token from the current scan buffer
+/* yylex: retrieve the next token from the current scan buffer
  * --------------
  *  return: an integer value representing the token type (enum TOKEN).  0 when
  *          no more input.
